/*
Changes in version 1.1.4:
- Widget dimensions now persistent between sessions.
- Object literals can now be used. For example: point = { x:0, y:6 }
- Toolbar: Use C instead of X

Changes in version 1.1.3:
- Fixed focus big introduced in 1.1.2

Changes in version 1.1.2:
- Numpad: The clear key clears the input area
- Numpad (International keyboards): the comma on the numpad is inserted as a period
  (tested on Safari 3.1, Leopard)
- Code cleanup based on latest version of jslint
  
Changes in version 1.1.1:
- Fixes for autocompletion when Safari 3.1 is installed.

Changes in version 1.1.0:
- Added autocompletion of variables and functions names using the Tab key. 
  For example, 1/an<TAB> autocompletes to 1/answer
- Added several new functions (log2, log10, and many trig & hyperbolic functions)
- Added degree mode.
- Fixed display and persistence of lists.
- Added several shortcut keys.
- Code cleanup based on Dustin Diaz's JavaScript work.

Changes in version 1.0.1:
- Default text is now "1+1" instead of "Enter an expression"
- Pi symbol is now set (type option-P to enter the pi character - Leopard only)
- Fixed minor layout issues in Leopard
- Widget now has a "back" with a "check version" link


Deployment steps:
[ ] Test on both 10.4 and 10.5
[ ] Update version number in Calc.html, Info.plist
[ ] Zip Calculate.wdgt and make sure it works with Stuffit
[ ] Update version number and docs at brackeen.com
[ ] Update at apple.com

User Requests:
[ ] Convert "a^b" to "pow(a,b)"
[ ] Locale-specific decimal separator for both input and output ("3,14159") - 
    impossible if lists/functions are allowed?

Rejected User Requests:
[-] Map the ^P and ^N [ctrl-P and ctrl-N] keys to 'cursor-up' and 'cursor-down'.
(That is also the default binding in bash, XCode, Emacs, TextEdit Pages, Mail and tons of other OS X programs ....)
    Rejected because it acts like "start of line" and "end of line" respectively.
    Also, there is no "cursor up/down" - there is only scrolling (different from the cursor)
Future Ideas:
[ ] Show memory: sort by name, type? (constants, variables, functions)
[ ] Show memory: delete button next to each var
[ ] Output: Hex, Decimal, or binary
[ ] Input as binary '0b11011'.
[ ] "Clear Memory" (with an undo option)
[ ] Save/restore input history.
[ ] matrix math functions
[ ] help in a separate html file?
[ ] Option: Do/Don't save memory
[ ] Regex to dissallow access of "window" object, other globals.
*/

var CalcWidget = window.CalcWidget || { };

/*
	Evaluate expressions in a "sandbox" so it doesn't have access to 
	the private members of CalcWidget.Calc or CalcWidget.UI, and so the user can't 
	override members of CalcWidget.Math.
*/
CalcWidget.evaluate = function(_expression, answer) {
	
	var sandbox = {};
	for (var i in CalcWidget.Math) {
		if (i.charAt(0) != '_') {
			sandbox[i] = CalcWidget.Math[i];
		}
	}
	with (sandbox) {
		// Unicode pi character (works only on Leopard)
		try {
			eval("var \u03c0 = Math.PI");
		}
		catch (ex) { }
		
		// Math-compatible
		var PI = Math.PI;
		var E = Math.E;
		
		return eval(_expression);
	}
};

/**
	Math functions
*/
CalcWidget.Math = {
	_angle: 1,

	// Constants
	pi:       Math.PI,
	e:        Math.E,
	
	// Basic Math 
	abs:      Math.abs,
	ceil:     Math.ceil,
	floor:    Math.floor,
	max:      Math.max,
	min:      Math.min,
	pow:      Math.pow,
	random:   Math.random,
	round:    Math.round,
	sqrt:     Math.sqrt,
	
	// Log functions
	exp:      Math.exp,
	log:      Math.log,
	ln:       Math.log,
	log10:    function(x) { return Math.log(x) / Math.log(10); },
	log2:     function(x) { return Math.log(x) / Math.log(2); },
	
	// Trig
	cos:      function(x) { x *= CalcWidget.Math._angle; return Math.cos(x); },
	sin:      function(x) { x *= CalcWidget.Math._angle; return Math.sin(x); },
	tan:      function(x) { x *= CalcWidget.Math._angle; return Math.tan(x); },
	cot:      function(x) { x *= CalcWidget.Math._angle; return 1/Math.tan(x); },
	sec:      function(x) { x *= CalcWidget.Math._angle; return 1/Math.cos(x); },
	csc:      function(x) { x *= CalcWidget.Math._angle; return 1/Math.sin(x); },
	
	acos:     function(x) { return Math.acos(x) / CalcWidget.Math._angle; },
	asin:     function(x) { return Math.asin(x) / CalcWidget.Math._angle; },
	atan:     function(x) { return Math.atan(x) / CalcWidget.Math._angle; },
	atan2:    function(x,y) { return Math.atan2(x,y) / CalcWidget.Math._angle; },
	acot:     function(x) { return Math.atan(1/x) / CalcWidget.Math._angle; },
	asec:     function(x) { return Math.acos(1/x) / CalcWidget.Math._angle; },
	acsc:     function(x) { return Math.asin(1/x) / CalcWidget.Math._angle; },
	
	// Hyperbolic functions
	sinh:     function(x) { x *= CalcWidget.Math._angle; return (Math.exp(x)-Math.exp(-x))/2; },
	cosh:     function(x) { x *= CalcWidget.Math._angle; return (Math.exp(x)+Math.exp(-x))/2; },
	tanh:     function(x) { x *= CalcWidget.Math._angle; return this.sinh(x) / this.cosh(x); },
	coth:     function(x) { x *= CalcWidget.Math._angle; return 1 / this.tanh(x); },
	sech:     function(x) { x *= CalcWidget.Math._angle; return 1 / this.cosh(x); },
	csch:     function(x) { x *= CalcWidget.Math._angle; return 1 / this.sinh(x); },
	
	asinh:    function(x) { return Math.log(x+Math.sqrt(x*x+1)) / CalcWidget.Math._angle; },
	acosh:    function(x) { return Math.log(x+Math.sqrt(x*x-1)) / CalcWidget.Math._angle; },
	atanh:    function(x) { return 0.5*Math.log((1+x)/(1-x)) / CalcWidget.Math._angle; },
	acoth:    function(x) { return 0.5*Math.log((x+1)/(x-1)) / CalcWidget.Math._angle; },
	asech:    function(x) { return Math.log(1/x+Math.sqrt(1/(x*x)+1)) / CalcWidget.Math._angle; },
	acsch:    function(x) { return Math.log(1/x+Math.sqrt(1/(x*x)-1)) / CalcWidget.Math._angle; }
};

CalcWidget.Calc = (function() {
		
	// Private
	
	// History of inputs
	var history = [];
	var historyIndex = -1;
	var maxHistory = 100;
	
	var lastAnswer = 0;
	var lastError = false;
	
	var knownMembers = [];
	
	function init() {
		// List of known variables at startup. Other variables will be considered "user variables"
		for (var i in window) {
			if (window.hasOwnProperty(i)) {
				knownMembers.push(i);
			}
		}
		loadUserVars();
	}
	
	function loadUserVars() {
		if (window.widget) {
			var memory = widget.preferenceForKey("memory");
			if (memory !== undefined && memory.length > 0) {
				try {
					CalcWidget.evaluate(memory, 0);
				}
				catch (ex) {
					// Do something?
				}
			}
		}
	}
	
	// Constructor
	init();
	
	// Public methods
	
	return {
	
		clearHistory: function() {
			history = [];
			historyIndex = -1;
		},
		
		getLastAnswer: function() {
			return lastAnswer;
		},
		
		getUserVars: function() {
			var userVars = [];
			
			for (var i in window) {
				if (window.hasOwnProperty(i)) {
					var isKnown = false;
					for (var j in knownMembers) {
						if (i === knownMembers[j]) {
							isKnown = true;
							break;
						}
					}
					if (!isKnown) {
						userVars.push(i);
					}
				}
			}
			
			return userVars;
		},
		
		saveUserVars: function() {
			var userVars = CalcWidget.Calc.getUserVars();
			var memory = "";
			for (var i in userVars) {
				if (userVars.hasOwnProperty(i)) {
					var name = userVars[i];
					if (window[name] !== undefined) {
                        var value = window[name];
                        if (typeof window[name] != "function") {
                            value = CalcWidget.valueToString(window[name]);
                        }

						memory += name + "=" + value + ";";
					}
				}
			}
            memory = memory.replace(/\n/g, ' ');
			if (window.widget) {
				widget.setPreferenceForKey(memory, "memory");
			}
		},
		
		clearUserVars: function() {
			lastAnswer = 0;
			lastError = false;
			
			var userVars = CalcWidget.Calc.getUserVars();
			for (var i in userVars) {
				if (userVars.hasOwnProperty(i)) {
					var name = userVars[i];
					try {
						eval("delete " + name);
					}
					catch (ex) {
						// Ignore
					}
				}
			}
			if (window.widget) {
				widget.setPreferenceForKey(null, "memory");
			}
		},
		
		calc: function(expression, addToHistory) {
			
			// Add to history
			if (addToHistory) {
				history.push(expression);
				if (history.length > maxHistory) {
					history = history.slice(history.length - maxHistory); 
				}
			}
			historyIndex = history.length;
				
			// Evaluate the expression
			var answer;
			lastError = false;
			try {
				answer = CalcWidget.evaluate(expression, lastAnswer);
				if (answer === undefined) {
					lastError = true;
				}
				else if (typeof answer == "number") {
					lastAnswer = answer;
				}
                else if (typeof answer == "function") {
                    answer = "Function defined";
                }
				else {
					answer = CalcWidget.valueToString(answer);
				}
			}
			catch (ex) {
				answer = ex;
				lastError = true;
			}
			
			return answer;
		},
		
		// Returns true if the last calculation in calc() resulted in an error
		wasError: function() {
			return lastError;
		},
		
		getHistoryPrev: function() {
			if (historyIndex > 0) { 
				historyIndex = Math.min(historyIndex - 1, history.length);
				return history[historyIndex];
			}
			else {
				return null;
			}
		},
		
		getHistoryNext: function() {
			if (historyIndex < history.length - 1) {
				historyIndex = Math.max(0, historyIndex + 1);
				return history[historyIndex];
			}
			else {
				historyIndex = history.length;
				return "";
			}
		}
	};
})();
